package polyglot.ext.ibex.lr;

import polyglot.ext.ibex.types.*;
import polyglot.util.*;
import java.util.*;
import polyglot.main.Report;

public class LRConstruction {
    final static boolean LALR = true;

    List nonterminals;
    List terminals;
    List rules;

    int[][] gotoTable;
    Set[][] actionTable;

    List[] edgeTable;
    List[] revEdgeTable;

    Collection startSymbols;
    Collection conflicts;

    /**
     * Map from source grammar start symbols to augmented grammar start states.
     */
    State[] startStatesMap;

    List startStates;

    List states;
    Map stateCache;
    boolean allowNewStates = true;

    GLRTerminal eofSymbol;

    public LRConstruction(Collection startSymbols, List terminals, List nonterminals, List rules, GLRTerminal eofSymbol) {
        this.startSymbols = startSymbols;
        this.terminals = terminals;
        this.nonterminals = nonterminals;
        this.rules = rules;
        this.eofSymbol = eofSymbol;

        allowNewStates = true;

        initStates();
        buildGotoStates();

        // Force an exception if we try to create any more states.
        allowNewStates = false;

        if (Report.should_report(TOPICS, 1)) {
            Report.report(1, "created " + states.size() + " states");
            Report.report(1, "    for " + nonterminals.size() + " nonterminals");
            Report.report(1, "        " + terminals.size() + " terminals");
            Report.report(1, "        " + rules.size() + " rules");
        }

        buildTables();
    }

    /** Initalize the state machine by creating the start states. */
    void initStates() {
        states = new ArrayList();
        stateCache = new HashMap();
        startStatesMap = new State[nonterminals.size()];
        startStates = new ArrayList();

        edgeTable = new List[terminals.size()+nonterminals.size()];
        revEdgeTable = new List[terminals.size()+nonterminals.size()];

        if (Report.should_report(TOPICS, 1))
            Report.report(1, "creating start states");

        for (Iterator i = startSymbols.iterator(); i.hasNext(); ) {
            GLRNonterminal s = (GLRNonterminal) i.next();

            for (Iterator j = s.rules().iterator(); j.hasNext(); ) {
                GLRRule r = (GLRRule) j.next();
                // Create a state for the start item.
                // As a side effect, it is added to the states set.
                State I = startState(r);
                startStatesMap[s.index()] = I;
                startStates.add(I);
            }
        }
    }

    /** Construct the goto states. */
    void buildGotoStates() {
        if (Report.should_report(TOPICS, 1))
            Report.report(1, "building states");

        // The start state(s) should already be in the list.
        LinkedList statesWorklist = new LinkedList(states);

        while (! statesWorklist.isEmpty()) {
            State I = (State) statesWorklist.removeFirst();

            for (Iterator j = I.items().iterator(); j.hasNext(); ) {
                Item item = (Item) j.next();

                GLRSymbol X = item.afterDot();

                // create a (possibly) new state
                State J = createGotoOrShiftState(I, X, statesWorklist);

                if (J != null) {
                    // Insert I ->^X J in the edge table and
                    // insert J ->^X I in the reverse edge table.

                    if (edgeTable[symbolIndex(X)] == null) {
                        edgeTable[symbolIndex(X)] = new ArrayList();
                    }

                    while (I.index() >= edgeTable[symbolIndex(X)].size()) {
                        edgeTable[symbolIndex(X)].add(null);
                    }

                    edgeTable[symbolIndex(X)].set(I.index(), J);

                    if (revEdgeTable[symbolIndex(X)] == null) {
                        revEdgeTable[symbolIndex(X)] = new ArrayList();
                    }

                    while (J.index() >= revEdgeTable[symbolIndex(X)].size()) {
                        revEdgeTable[symbolIndex(X)].add(null);
                    }

                    Collection l = (Collection) revEdgeTable[symbolIndex(X)].get(J.index());
                    if (l == null) {
                        l = new HashSet();
                        revEdgeTable[symbolIndex(X)].set(J.index(), l);
                    }

                    l.add(I);
                }
            }
        }
    }

    /** Build the action and goto tables. */
    void buildTables() {
        if (Report.should_report(TOPICS, 1))
            Report.report(1, "building tables");

        conflicts = new HashSet();

        actionTable = new Set[states.size()][terminals.size()];
        gotoTable = new int[states.size()][nonterminals.size()];

        for (int i = 0; i < actionTable.length; i++) {
            for (int j = 0; j < actionTable[i].length; j++) {
                actionTable[i][j] = new HashSet();
            }
        }

        for (int i = 0; i < states.size(); i++) {
            State I = (State) states.get(i);

            for (Iterator j = I.items().iterator(); j.hasNext(); ) {
                Item item = (Item) j.next();

                GLRSymbol X = item.afterDot();

                if (X == null) {
                    for (Iterator k = item.lookahead.iterator(); k.hasNext(); ) {
                        GLRTerminal t = (GLRTerminal) k.next();
                        addAction(I, t, new Reduce(item.rule));
                    }
                }
                else if (X.equals(eofSymbol)) {
                    addAction(I, eofSymbol, new Accept(item.rule.lhs()));
                }
                else {
                    State J = gotoOrShiftState(I, X);

                    if (J == null) {
                        throw new InternalCompilerError(
                            "Unreachable next state for " + I + ": " + item);
                    }

                    // if J is a new state, it will be appended to the
                    // states list and we'll process it soon

                    if (X instanceof GLRNonterminal) {
                        setGoto(I, (GLRNonterminal) X, J);
                    }
                    else {
                        addAction(I, (GLRTerminal) X,
                                  new Shift(J, ((GLRTerminal) X)));
                    }
                }
            }
        }

        if (Report.should_report(TOPICS, 1)) {
            Report.report(1, "done building tables");
            if (conflicts.size() > 0) {
                Report.report(1, conflicts.size() + " LR conflicts found:");
                int k = 0;
                for (Iterator i = conflicts.iterator(); i.hasNext(); ) {
                    Conflict c = (Conflict) i.next();
                    Report.report(1, " Conflict #" + (++k));
                    c.dump();
                }
            }
        }
    }

    Collection conflicts() {
        return conflicts;
    }

    Set actions(State I, GLRTerminal t) {
        return actionTable[I.index()][t.index()];
    }

    void addAction(State I, GLRTerminal t, Action a) {
        Set c = actions(I, t);
        c.add(a);
        if (c.size() > 1) {
            conflicts.add(new Conflict(I, t, this));
        }
    }

    State gotoState(State I, GLRNonterminal t) {
        return (State) states.get(gotoTable[I.index()][t.index()]);
    }

    void setGoto(State I, GLRNonterminal t, State J) {
        gotoTable[I.index()][t.index()] = J.index();
    }

    State gotoOrShiftState(State I, GLRSymbol X) {
        if (X != null && edgeTable[symbolIndex(X)] != null &&
            I.index() < edgeTable[symbolIndex(X)].size()) {
            return (State) edgeTable[symbolIndex(X)].get(I.index());
        }

        return null;
    }

    Collection reverseGotoOrShiftStates(State I, GLRSymbol X) {
        if (X != null && revEdgeTable[symbolIndex(X)] != null &&
            I.index() < revEdgeTable[symbolIndex(X)].size()) {
            Collection c = (Collection) revEdgeTable[symbolIndex(X)].get(I.index());
            if (c != null) return c;
        }

        return Collections.EMPTY_LIST;
    }

    void dump() {
        for (int i = 0; i < terminals.size(); i++) {
            GLRTerminal t = (GLRTerminal) terminals.get(i);
            System.out.println("Terminal " + i + ": " + t);
        }
        System.out.println();

        for (int i = 0; i < nonterminals.size(); i++) {
            GLRNonterminal t = (GLRNonterminal) nonterminals.get(i);
            System.out.println("Nonterminal " + i + ": " + t);
        }
        System.out.println();

        for (int i = 0; i < rules.size(); i++) {
            GLRRule t = (GLRRule) rules.get(i);
            System.out.println("Rule " + t);
        }
        System.out.println();

        for (int i = 0; i < states.size(); i++) {
            State I = (State) states.get(i);

            I.dump();

            System.out.println();

            int[] gotoRow = gotoTable[I.index()];

            for (int j = 0; j < gotoRow.length; j++) {
                // Assume we never goto state 0.
                // This should be true since state 0 is a start state
                // and any reductions of the initial rule in the
                // state will transition to an accept state.
                //
                // This isn't important anyway since we're just dumping
                // debug output.
                if (gotoRow[j] != 0) {
                    System.out.println("    on " + nonterminals.get(j) +
                                       ", goto " + states.get(gotoRow[j]));
                }
            }

            Set[] actionRow = actionTable[I.index()];

            for (int j = 0; j < actionRow.length; j++) {
                if (! actionRow[j].isEmpty()) {
                    System.out.println("    on " + terminals.get(j) +
                                       ", " + actionRow[j]);
                }
            }

            System.out.println();
        }
    }

    int symbolIndex(GLRSymbol X) {
        if (X instanceof GLRTerminal) {
            return X.index() + nonterminals.size();
        }
        return X.index();
    }

    Item startItem(GLRRule rule) {
        return new Item(rule, 0, Collections.EMPTY_SET);
    }

    State startState(GLRRule rule) {
        return closure(Collections.singleton(startItem(rule)), null);
    }

    State createStateForItemSet(Set items, List statesWorklist) {
        State s = (State) stateCache.get(items);

        if (s == null) {
            s = new State(states.size(), items);
            states.add(s);
            stateCache.put(items, s);

            if (statesWorklist != null) {
                statesWorklist.add(s);
            }
        }
        else if (LALR) {
            // LALR(1): merge items
            boolean changed = false;

            for (Iterator i = items.iterator(); i.hasNext(); ) {
                Item item = (Item) i.next();
                for (Iterator j = s.items.iterator(); j.hasNext(); ) {
                    Item sitem = (Item) j.next();
                    if (item.equals(sitem)) {
                        changed |= sitem.lookahead.addAll(item.lookahead);
                    }
                }
            }

            if (changed && statesWorklist != null) {
                statesWorklist.add(s);
            }
        }

        return s;
    }

    State createGotoOrShiftState(State I, GLRSymbol X, List statesWorklist) {
        if (X == null || X.equals(eofSymbol)) {
            return null;
        }

        Set J = new HashSet();
        for (Iterator i = I.items().iterator(); i.hasNext(); ) {
            Item item = (Item) i.next();
            if (item.afterDot() != null && item.afterDot().equals(X)) {
                J.add(item.shiftDot());
            }
        }

        if (J.isEmpty()) {
            return null;
        }

        return closure(J, statesWorklist);
    }

    State closure(Set items, List statesWorklist) {
        ArrayList worklist = new ArrayList(items);
        Set closure = new HashSet(items);

        while (worklist.size() > 0) {
            Item item = (Item) worklist.remove(worklist.size()-1);

            // item == (A -> alpha . X beta, z)
            GLRSymbol X = item.afterDot();

            List beta = new ArrayList();

            if (item.dot < item.rule.rhs().size()) {
                beta.addAll(item.rule.rhs().subList(item.dot+1, item.rule.rhs().size()));
            }

            Set first = first(beta, item.lookahead);

            if (X instanceof GLRNonterminal) {
                // item == (A -> alpha . X beta, z)
                List rules = ((GLRNonterminal) X).rules();

                for (Iterator j = rules.iterator(); j.hasNext(); ) {
                    GLRRule r = (GLRRule) j.next();
                    // r == X -> gamma
                    boolean changed = false;
                    boolean found = false;

                    List addToClosure = new ArrayList();

                    for (Iterator k = closure.iterator(); k.hasNext(); ) {
                        Item ritem = (Item) k.next();
                        if (ritem.rule == r && ritem.dot == 0) {
                            found = true;
                            changed |= ritem.lookahead.addAll(first);

                            if (changed) {
                                // The hash code may have changed; remove from
                                // the closure, to add back after the iterator
                                // finishes.  If we don't remove and add back,
                                // the item could be in the wrong hash
                                // bucket and won't be found.
                                k.remove();
                                addToClosure.add(ritem);
                            }
                        }
                    }

                    if (! found) {
                        Item ritem = new Item(r, 0, first);
                        closure.add(ritem);
                        worklist.add(ritem);
                    }

                    closure.addAll(addToClosure);
                    worklist.addAll(addToClosure);
                }
            }
        }

        return createStateForItemSet(closure, statesWorklist);
    }

    static Set first(List beta, Set z) {
        Set first = new HashSet();

        for (Iterator i = beta.iterator(); i.hasNext(); ) {
            GLRSymbol s = (GLRSymbol) i.next();
            first.addAll(s.first());
            if (! s.isNullable()) {
                return first;
            }
        }

        first.addAll(z);

        return first;
    }

    static Collection TOPICS = Arrays.asList( new String[] { "lr", "ibex" });
}
