package polyglot.ext.ibex.lr;

import polyglot.ext.ibex.types.*;
import polyglot.util.*;
import java.util.*;
import polyglot.main.Report;

public class MergeDetection {
    List nonterminals;
    List terminals;
    List rules;
    List states;
    GLR glr;

    /**
     * Map from nonterminals pairs of rules requiring merge actions.
     */
    Set[] mergePairs;

    public MergeDetection(GLR glr)
    {
        this.glr = glr;
        this.states = glr.lrc.states;
        this.nonterminals = glr.nonterminals;
        this.terminals = glr.terminals;
        this.rules = glr.rules;

        computeMergePoints();
    }

    /**
     * Return a collection of collections of right-hand-sides
     * <code>rhs</code> associated with the nonterminal <code>lhs</code>
     * such that the rule <code>lhs ::= rhs</code> requires a merge
     * action.  Each pair of rules in each collection should have a merge
     * action.
     */
    public Collection mergeRulesForNonterminal(GLRNonterminal n) {
        Set pairs = mergePairs[n.index()];
        
        if (pairs == null) {
            return Collections.EMPTY_LIST;
        }

        ArrayList groups = new ArrayList();

        for (Iterator i = pairs.iterator(); i.hasNext(); ) {
            ReducePair p = (ReducePair) i.next();
            List pair = new ArrayList(2);
            pair.add(p.r1.nontermRhs);
            pair.add(p.r2.nontermRhs);
            groups.add(pair);
        }

        return groups;
    }

    void buildReachableSets(BitSet[] reachableFrom) {
        for (int i = 0; i < glr.nonterminals.size(); i++) {
            reachableFrom[i] = new BitSet();
            reachableFrom[i].set(i);
        }

        for (Iterator i = glr.lrc.startSymbols.iterator(); i.hasNext(); ) {
            GLRNonterminal s_ = (GLRNonterminal) i.next();
            reachableDFS(s_, reachableFrom);
        }
    }

    void reachableDFS(GLRNonterminal s, BitSet[] reachableFrom) {
        for (Iterator i = s.rules().iterator(); i.hasNext(); ) {
            GLRRule r = (GLRRule) i.next();

            for (Iterator j = r.rhs().iterator(); j.hasNext(); ) {
                GLRSymbol X = (GLRSymbol) j.next();
                if (X instanceof GLRNonterminal) {
                    BitSet oldR = reachableFrom[X.index()];
                    BitSet newR = new BitSet();
                    newR.or(oldR);
                    newR.or(reachableFrom[s.index()]);
                    if (! oldR.equals(newR)) {
                        reachableFrom[X.index()] = newR;
                        reachableDFS((GLRNonterminal) X, reachableFrom);
                    }
                }
            }
        }
    }

    /**
     * Compute the ambiguous paths in the shift-reduce graph.
     *
     * When a state has two or more conflicting actions for a given
     * terminal, the parser performs all the actions, following multiple
     * paths through the state machine.  Eventually, either all but one
     * path will fail to shift a terminal and be discarded, or, if the
     * grammar is truly ambiguous, the paths will converge on an edge
     * for some nonterminal.
     *
     * This function returns a list of shortest paths from a state with
     * a conflict to a merging state.
     *
     * A path is represented by the reductions that occur on the path.
     *
     * Merge action are required when two or more paths end in the same
     * state.  The merge action is keyed on the last reduction in each
     * path.
     */
    void computeMergePoints() {

        BitSet[] reachableFrom = new BitSet[nonterminals.size()];
        buildReachableSets(reachableFrom);

        // Now find all states that have a shift/reduce or reduce/reduce
        // conflict.
        List conflictStates = new ArrayList();

        for (Iterator i = glr.lrc.conflicts().iterator(); i.hasNext(); ) {
            Conflict c = (Conflict) i.next();
            conflictStates.add(c.state());
        }

        if (conflictStates.isEmpty()) {
            if (Report.should_report(TOPICS, 1)) {
                Report.report(2, "no conflicts");
            }
            return;
        }

        // Compute may-merge rules.
        List reducePairs = new ArrayList();

        // First build the shift-reduce graph.  This is the graph formed
        // by taking the shift edges from the parser state machine and
        // augmenting it with the edges formed by simulating a reduce
        // and state change.
        SRGraph srGraph = new SRGraph(glr.lrc);

        if (Report.should_report(TOPICS, 2)) {
            Report.report(2, "building DFA from SR graph");
        }

        // Build a DFA from the SR graph.
        Collection startStates = srGraph.nodesForParserStates(conflictStates);
        SRDFA dfa = new SRDFA(srGraph, startStates);

        rhsIntersectCache = new int[rules.size()][rules.size()];

        for (int i = 0; i < conflictStates.size(); i++) {
            State I = (State) conflictStates.get(i);

            if (Report.should_report(TOPICS, 2)) {
                Report.report(2, "conflict at " + I);
                I.dump();
            }

            for (Iterator nodes = srGraph.nodesForParserState(I).iterator();
                nodes.hasNext(); ) {

                SRNode n = (SRNode) nodes.next();

                if (Report.should_report(TOPICS, 2)) {
                    Report.report(2, "building G(" + n + ")");
                }

                RootedGraph g = new RootedGraph(n, srGraph);

                if (Report.should_report(TOPICS, 2)) {
                    Report.report(2, "computing dominators for G(" + n + ")");
                }

                // Now compute the dominance frontier of the successors
                // of the conflict node.
                Dominators dom = new Dominators(g);
                BitSet DF = new BitSet();

                for (Iterator j = g.succs(g.root()); j.hasNext(); ) {
                    Node k = ((SREdge) j.next()).target();
                    BitSet DFj = dom.DF(k);
                    DF.or(DFj);
                }

                // We need only consider states reachable from the DF
                // that contain items for nonterminals that can
                // derive the conflict.
                BitSet reachableFromDF = new BitSet();

                // Collect states reachable from the dom frontier. 
                for (int j = DF.nextSetBit(0); j >= 0; j = DF.nextSetBit(j+1)) {
                    dfs(g, g.root(), reachableFromDF, reachableFrom);
                }

                if (Report.should_report(TOPICS, 6)) {
                    Report.report(6, "DF = " + DF);
                    Report.report(6, "reachable from DF = " + reachableFromDF);
                }

                // For each reachable state with more than one incoming reduce
                // edge, check if the state is reachable by more than one path
                // from the root containing the same shifts
                for (int j = reachableFromDF.nextSetBit(0); j >= 0;
                    j = reachableFromDF.nextSetBit(j+1)) {
                    SRNode jn = srGraph.nodeAt(j);
                    checkPaths(g, jn, reducePairs, dfa);
                }
            }
        }

        // Free up some space.
        srGraph = null;
        dfa = null;

        if (! reducePairs.isEmpty()) {
            // Build an approximation FA.
            // XXX should build one machine covering all start symbols,
            // not one machine per symbol.
            ApproximationFA afa = null;
            Partition partitions = null;

            for (Iterator i = glr.lrc.startSymbols.iterator(); i.hasNext(); ) {
                GLRNonterminal s_ = (GLRNonterminal) i.next();
                partitions = new Partition(nonterminals, s_);
                afa = new ApproximationFA(terminals, nonterminals, s_, partitions);
                break;
            }

            if (afa == null) {
                throw new InternalCompilerError("no start symbol");
            }

            for (Iterator i = reducePairs.iterator(); i.hasNext(); ) {
                ReducePair p = (ReducePair) i.next();
                if (! slowReducePairIntersect(afa, p)) {
                    i.remove();
                }
            }
        }

        mergePairs = new Set[nonterminals.size()];

        for (Iterator i = reducePairs.iterator(); i.hasNext(); ) {
            ReducePair p = (ReducePair) i.next();
            if (mergePairs[p.r1.lhs.index()] == null) {
                mergePairs[p.r1.lhs.index()] = new HashSet();
            }
            mergePairs[p.r1.lhs.index()].add(p);
        }
    }

    void checkPaths(RootedGraph g, SRNode j, List reducePairs, SRDFA dfa) {
        List reduces = new ArrayList();

        if (Report.should_report(TOPICS, 2)) {
            Report.report(2, "checking paths to " + j);
        }

        for (Iterator si = g.preds(j); si.hasNext(); ) {
            // Forward edge from i to j
            RevEdge ji = (RevEdge) si.next();
            SRNode i = (SRNode) ji.target();

            SREdge ij = (SREdge) ji.forward();

            if (ij instanceof ReduceEdge) {
		// Examine the other reduce edges into j to see if there
		// is another path to j.
                reduces.add(ij);
            }
        }

        for (int ri = 0; ri < reduces.size(); ri++) {
            // Forward edge from i to j.
            ReduceEdge ij = (ReduceEdge) reduces.get(ri);
            SRNode i = (SRNode) ij.source();

            for (int rk = ri+1; rk < reduces.size(); rk++) {
                // Forward edge from k to j.
                ReduceEdge kj = (ReduceEdge) reduces.get(rk);
                SRNode k = (SRNode) kj.source();

                if (k.index() == i.index()) {
                    continue;
                }

                if (Report.should_report(TOPICS, 2)) {
                    Report.report(2, "checking " + ij);
                    Report.report(2, " against " + kj);
                }

                ReducePair p = new ReducePair(
                    ((ReduceEdge) ij).reduction, kj.reduction);

                if (reducePairs.contains(p)) {
                    if (Report.should_report(TOPICS, 2)) {
                        Report.report(2, "redundant");
                    }
                    continue;
                }

                if (fastReducePairIntersect(p)) {
                    // (j,i) is a reduce edge, and
                    // (k,i) is a reduce edge.
                    // Check if both are reachable via the same set of shifts.
                    if (dfa.intersectionNonEmpty((SRNode) g.root(), k, i)) {
                        if (Report.should_report(TOPICS, 3)) {
                            Report.report(2, "definite merge at " + p);
                        }
                        reducePairs.add(p);
                    }
                    else {
                        if (Report.should_report(TOPICS, 3)) {
                            Report.report(3, "not a merge: paths do not merge at " + j);
                        }
                    }
                }
            }
        }
    }

    boolean slowReducePairIntersect(ApproximationFA afa, ReducePair p) {
        GLRRule r1 = p.r1;
        GLRRule r2 = p.r2;

        if (rhsIntersectCache[r1.index][r2.index] == 1) {
            return true;
        }

        if (rhsIntersectCache[r1.index][r2.index] == -1) {
            return false;
        }

        if (Report.should_report(TOPICS, 2)) {
            Report.report(2, "slow check for merge at " + p);
        }

        boolean intersect = true;

        if (! rhsApproxIntersects(afa, r1, r2)) {
            if (Report.should_report(TOPICS, 3)) {
                Report.report(3, "not a merge: rhs do not intersect");
            }
            intersect = false;
        }

        rhsIntersectCache[r1.index][r2.index] = intersect ? 1 : -1;
        rhsIntersectCache[r2.index][r1.index] = intersect ? 1 : -1;

        return intersect;
    }

    boolean fastReducePairIntersect(ReducePair p) {
        if (Report.should_report(TOPICS, 2)) {
            Report.report(2, "possible merge at " + p);
        }

        GLRRule r1 = p.r1;
        GLRRule r2 = p.r2;

        if (r1.lhs.index() != r2.lhs.index()) {
            if (Report.should_report(TOPICS, 3)) {
                Report.report(3, "not a merge: lhs do not match");
            }
            return false;
        }

        // Check if the RHS of the rules intersect.
        if (rhsIntersectCache[r1.index][r2.index] == 1) {
            if (Report.should_report(TOPICS, 3)) {
                Report.report(3, "not a merge: (cached) rhs intersect");
            }
            return true;
        }

        // Check if the RHS of the rules intersect.
        if (rhsIntersectCache[r1.index][r2.index] == -1) {
            if (Report.should_report(TOPICS, 3)) {
                Report.report(3, "not a merge: (cached) rhs do not intersect");
            }
            return false;
        }

        if (r1.index() == r2.index()) {
            // Same rule.  Check if a conflict is derivable.
            // XXX: the check is slow: just say it is!
            if (true || rhsConflicts(r1)) {
                if (Report.should_report(TOPICS, 3)) {
                    Report.report(3, "identity merge at " + p);
                }

                rhsIntersectCache[r1.index()][r2.index()] = 1;
                return true;
            }
            else {
                if (Report.should_report(TOPICS, 3)) {
                    Report.report(3, "not a merge: same rule, but not ambiguous");
                }

                rhsIntersectCache[r1.index()][r2.index()] = -1;
                return false;
            }
        }

        boolean intersect = true;

        // First, do a quick test of the first sets of each rule.
        // Then, compute the LR states for each rule and see
        // if they conflict.
        // Then, build approximation DFAs for each rule
        // and see if they intersect.
        if (! firstSetsIntersect(r1, r2)) {
            if (Report.should_report(TOPICS, 3)) {
                Report.report(3, "not a merge: first sets do not intersect");
            }
            intersect = false;
        }
        else if (! leftRecursionCheck(r1, r2)) {
            if (Report.should_report(TOPICS, 3)) {
                Report.report(3, "not a merge: left recursion pattern match");
            }
            intersect = false;
        }
        else if (! rightRecursionCheck(r1, r2)) {
            if (Report.should_report(TOPICS, 3)) {
                Report.report(3, "not a merge: right recursion pattern match");
            }
            intersect = false;
        }

        // Cache negative results, but not positive results since we
        // need to do more checking later.
        if (intersect == false) {
            rhsIntersectCache[r1.index][r2.index] = -1;
            rhsIntersectCache[r2.index][r1.index] = -1;
        }

        return intersect;
    }

    static class ReducePair {
        GLRRule r1, r2;

        ReducePair(GLRRule r1, GLRRule r2) {
            if (r1.index() <= r2.index()) {
                this.r1 = r1;
                this.r2 = r2;
            }
            else {
                this.r1 = r2;
                this.r2 = r1;
            }
        }

        public String toString() {
            return "[" + r1 + ", " + r2 + "]";
        }

        public int hashCode() {
            return r1.hashCode() + r2.hashCode();
        }

        public boolean equals(Object o) {
            return o instanceof ReducePair
                && r1.equals(((ReducePair) o).r1)
                && r2.equals(((ReducePair) o).r2);
        }
    }

    int[][] rhsIntersectCache;

    boolean leftRecursionCheck(GLRRule r1, GLRRule r2) {
        // A quick test for the very common case:
        //
        // A -> alpha
        // A -> A beta alpha
        //
        // return false if beta is not nullable
        // return false if alpha is not nullable
        // return true otherwise
        //
        // XXX: This is probably unsound!

        if (r1.lhs.rules.size() != 2) {
            return true;
        }

        // Make the shorter rule first
        if (r1.rhs.size() > r2.rhs.size()) {
            GLRRule tmp = r1;
            r1 = r2;
            r2 = tmp;
        }

        System.out.println("left recursion test: r1 = " + r1);
        System.out.println("left recursion test: r2 = " + r2);

        if (r1.rhs.size() == 0) {
            return true;
        }

        GLRNonterminal A = r1.lhs;

        GLRSymbol X1 = (GLRSymbol) r1.rhs.get(0);
        GLRSymbol X2 = (GLRSymbol) r2.rhs.get(0);

        // Ensure r1 does not start with A
        if (X1.equals(A)) {
            return true;
        }

        // Ensure r2 does start with A
        if (! X2.equals(A)) {
            return true;
        }

        System.out.println("possible left recursive rule pair " + r1 + " and " + r2);

        // Ensure r2 ends with r1.rhs
        int diff = r2.rhs.size() - r1.rhs.size();

        for (int i = 0; i < r1.rhs.size(); i++) {
            X1 = (GLRSymbol) r1.rhs.get(i);
            X2 = (GLRSymbol) r2.rhs.get(i + diff);
            if (! X1.equals(X2)) {
                return true;
            }
        }

        System.out.println("left recursive rule pair " + r1 + " and " + r2);

        boolean nullable;

        // Check if beta alpha is nullable.
        nullable = true;

        for (int i = 1; i < r2.rhs.size(); i++) {
            X2 = (GLRSymbol) r2.rhs.get(i);
            if (! X2.isNullable()) {
                nullable = false;
                break;
            }
        }

        if (! nullable) {
            return false;
        }

        return true;
    }

    boolean rightRecursionCheck(GLRRule r1, GLRRule r2) {
        // A quick test for the very common case:
        //
        // A -> alpha
        // A -> alpha beta A
        //
        // return false if beta is not nullable
        // return false if alpha is not nullable
        // return true otherwise
        //
        // XXX: This is probably unsound!

        if (r1.lhs.rules.size() != 2) {
            return true;
        }

        // Make the shorter rule first
        if (r1.rhs.size() > r2.rhs.size()) {
            GLRRule tmp = r1;
            r1 = r2;
            r2 = tmp;
        }

        System.out.println("right recursion test: r1 = " + r1);
        System.out.println("right recursion test: r2 = " + r2);

        if (r1.rhs.size() == 0) {
            return true;
        }

        GLRNonterminal A = r1.lhs;

        GLRSymbol X1 = (GLRSymbol) r1.rhs.get(0);
        GLRSymbol X2 = (GLRSymbol) r2.rhs.get(r2.rhs.size()-1);

        // Ensure r1 does not start with A
        if (X1.equals(A)) {
            return true;
        }

        // Ensure r2 ends with A
        if (! X2.equals(A)) {
            return true;
        }

        System.out.println("possible right recursive rule pair " + r1 + " and " + r2);

        // Ensure r2 starts with r1.rhs
        for (int i = 0; i < r1.rhs.size(); i++) {
            X1 = (GLRSymbol) r1.rhs.get(i);
            X2 = (GLRSymbol) r2.rhs.get(i);
            if (! X1.equals(X2)) {
                return true;
            }
        }

        System.out.println("right recursive rule pair " + r1 + " and " + r2);

        boolean nullable;

        // Check if alpha beta is nullable.
        nullable = true;

        for (int i = 0; i < r2.rhs.size() - 1; i++) {
            X2 = (GLRSymbol) r2.rhs.get(i);
            if (! X2.isNullable()) {
                nullable = false;
                break;
            }
        }

        if (! nullable) {
            return false;
        }

        return true;
    }

    boolean firstSetsIntersect(GLRRule r1, GLRRule r2) {
        // Don't pass in the follow sets.  We're just trying to
        // see of the strings derivable from r1 and r2
        // intersect.  The first set is an approximation.
        Set f1 = glr.lrc.first(r1.rhs, Collections.EMPTY_SET);
        Set f2 = glr.lrc.first(r2.rhs, Collections.EMPTY_SET);

        boolean intersect = false;

        if (f1.isEmpty() != f2.isEmpty()) {
            // One is an epsilon rule, the other isn't.
            // No intersection!
            intersect = false;
        }
        else {
            Set intersection = new HashSet(f1);
            intersection.retainAll(f2);
            intersect = ! intersection.isEmpty();
        }

        return intersect;
    }

    void dfs(RootedGraph g, Node i, BitSet visited, BitSet[] reachableFrom) {
        if (visited.get(i.index())) {
            return;
        }

        visited.set(i.index());

        for (Iterator si = g.succs(i); si.hasNext(); ) {
            // Edge from i to j
            SREdge ij = (SREdge) si.next();

            boolean followEdge = true;

            if (ij instanceof ReduceEdge) {
                GLRRule r = ((ReduceEdge) ij).reduction;

                // Check if the nonterminal we're reducing to
                // is an ancestor of any of the nonterminals in the
                // conflicting state.
                //
                // FIXME: refine to only include check items that actually
                // conflict.
                State I = ((SRNode) g.root()).state();

                for (Iterator j = I.items.iterator(); j.hasNext(); ) {
                    Item item = (Item) j.next();

                    if (! reachableFrom[item.rule.lhs.index()].get(r.lhs.index())) {
                        followEdge = false;
                        break;
                    }
                }
            }

            if (followEdge) {
                dfs(g, ij.target(), visited, reachableFrom);
            }
        }
    }

    boolean rhsConflicts(GLRRule r1, GLRRule r2) {
        // Want to see if S -> alpha and S -> beta conflict

        // Build the LR parser with start rule A -> B $
        //                           and rules B -> alpha
        //                                 and B -> beta

        List nonterminals = new ArrayList(this.nonterminals.size()+2);
        nonterminals.addAll(this.nonterminals);

        List rules = new ArrayList(this.rules.size()+3);
        rules.addAll(this.rules);

        GLRNonterminal A = new GLRNonterminal(r1.lhs.nonterm,
                                              r1.lhs.nonterm.name() +
                                              "!rhs!start",
                                              nonterminals.size());
        nonterminals.add(A);

        GLRNonterminal B = new GLRNonterminal(r1.lhs.nonterm,
                                              r1.lhs.nonterm.name() + "!rhs",
                                              nonterminals.size());
        nonterminals.add(B);

        List l0 = new ArrayList(2);
        l0.add(B);
        l0.add(glr.eofSymbol());

        List l1 = r1.rhs;
        List l2 = r2.rhs;

        GLRRule r0_ = new GLRRule(null, null, A, l0, rules.size());
        rules.add(r0_);
        A.rules.add(r0_);

        GLRRule r1_ = new GLRRule(null, null, B, l1, rules.size());
        rules.add(r1_);
        B.rules.add(r1_);

        GLRRule r2_ = new GLRRule(null, null, B, l2, rules.size());
        rules.add(r2_);
        B.rules.add(r2_);

        LRConstruction lrc = new LRConstruction(
            Collections.singleton(A), terminals, nonterminals,
            rules, glr.eofSymbol());

        if (Report.should_report(TOPICS, 4)) {
            Report.report(4, "LR machine for " + r1 + " and " + r2);
            lrc.dump();
        }

        // Is there a conflict?
        Collection conflicts = lrc.conflicts();

        boolean result = ! conflicts.isEmpty();

        if (Report.should_report(TOPICS, 4)) {
            Report.report(4, r1 + " and " + r2 + " conflict = " + result);
        }

        return result;
    }

    boolean rhsConflicts(GLRRule r) {
        // Want to see if S -> alpha on its own is ambiguouon its own is
        // ambiguous

        // Build the LR parser with start rule A -> alpha $

        List nonterminals = new ArrayList(this.nonterminals.size()+1);
        nonterminals.addAll(this.nonterminals);

        List rules = new ArrayList(this.rules.size()+1);
        rules.addAll(this.rules);

        GLRNonterminal A = new GLRNonterminal(r.lhs.nonterm,
                                              r.lhs.nonterm.name() +
                                              "!rhs!start",
                                              nonterminals.size());
        nonterminals.add(A);

        List l = new ArrayList(2);
        l.addAll(r.rhs);
        l.add(glr.eofSymbol());

        GLRRule r_ = new GLRRule(null, null, A, l, rules.size());
        rules.add(r_);
        A.rules.add(r_);

        LRConstruction lrc = new LRConstruction(
            Collections.singleton(A), terminals, nonterminals,
            rules, glr.eofSymbol());

        if (Report.should_report(TOPICS, 4)) {
            Report.report(4, "LR machine for " + r);
            lrc.dump();
        }

        // Is there a conflict?
        Collection conflicts = lrc.conflicts();

        boolean result = ! conflicts.isEmpty();

        if (Report.should_report(TOPICS, 4)) {
            Report.report(4, r + " ambiguous = " + result);
        }

        return result;
    }

    List rhs(GLRRule r) {
        // Expand unit productions (A -> B).
        while (r.rhs.size() == 1) {
            GLRSymbol X = (GLRSymbol) r.rhs.get(0);
            if (X instanceof GLRNonterminal) {
                GLRNonterminal A = (GLRNonterminal) X;
                if (A.rules().size() == 1) {
                    r = (GLRRule) A.rules().get(0);
                    continue;
                }
            }

            break;
        }

        return r.rhs;
    }

    boolean rhsApproxIntersects(ApproximationFA afa, GLRRule r1, GLRRule r2) {
        List l1 = rhs(r1);
        List l2 = rhs(r2);

        boolean result = afa.intersects(l1, l2);

        if (Report.should_report(TOPICS, 3)) {
            Report.report(3, "rhs intersect = " + result);
        }

        return result;
    }

    // Do s1 and s2 have the same set of conflicts?
    boolean sameActions(Collection s1, Collection s2) {
        for (Iterator i1 = s1.iterator(); i1.hasNext(); ) {
            Action a1 = (Action) i1.next();

            boolean found = false;

            for (Iterator i2 = s2.iterator(); i2.hasNext(); ) {
                Action a2 = (Action) i2.next();

                if (a1 instanceof Reduce && a2 instanceof Reduce &&
                    ((Reduce) a1).rule.equals(((Reduce) a2).rule)) {
                    found = true;
                    break;
                }
                if (a1 instanceof Shift && a2 instanceof Shift) {
                    found = true;
                    break;
                }
                if (a1 instanceof Accept && a2 instanceof Accept) {
                    found = true;
                    break;
                }
            }

            if (! found) {
                return false;
            }
        }

        return true;
    }

    static Collection TOPICS = Arrays.asList( new String[] { "lr", "ibex", "merge" });
}
