package polyglot.ext.polyj.ast;

import polyglot.ext.polyj.extension.*;
import polyglot.ext.polyj.types.*;
import polyglot.ext.polyj.util.*;
import polyglot.ext.polyj.visit.*;

import polyglot.ext.jl.ast.*;

import polyglot.ast.*;
import polyglot.types.*;
import polyglot.util.*;
import polyglot.visit.*;

import java.util.*;

/**
 * A <code>PolyJCall</code> is an immutable representation of a Java
 * method call.  It consists of a method name and a list of arguments.
 * It may also have either a Type upon which the method is being
 * called or an expression upon which the method is being called.
 */
public class PolyJCall_c extends Call_c implements PolyJCall {

    /**
     * Create a new <code>PolyJCall</code>.
     *
     * @pre Requires that <code>target</code> is either a <code>TypeNode</code>
     *  or an <code>Expr</code> or <code>null</code>. Also
     *  <code>arguments</code> must be a list of <code>Expr</code>s.
     */
    public PolyJCall_c( Position pos, Receiver target, String name, List arguments) {
        super(pos, target, name, arguments);
    }

    /**
     * Where methods, static parametric methods, and primitive types
     * (boxing/unboxing) must be translated differently than in Java.
     */
    public void translate(CodeWriter w, Translator tr) {
        Context c = tr.context();
        PolyJTypeSystem ts = (PolyJTypeSystem) c.typeSystem();
        MethodInstance myMi = methodInstance();
        boolean isValueIgnored = ((PolyJExt) this.ext()).isValueIgnored();
	ReturnTypeMutator m = new ReturnTypeMutator(myMi);
	boolean insertCast = m.typeChanges() && !isValueIgnored;

	//insertCast = false;
	if (insertCast) {
	    w.write("((");
	    w.write(m.castTo().translate(c));
	    w.write(") ");
	}

        if (target != null && target.type() instanceof ParamType
                && !ts.equals(myMi.container(),(ts.Object()))) {
            PolyJTranslator.translateWhereMethod(w, tr, ts, c, 
                (ParamType) target.type(), this, target, myMi, arguments());
        } else if (target != null
                && myMi.container() instanceof InstType) {
            translateInstMethod(w, tr, ts, c);
        } else {
            translateDefault(w, tr);
        }

	if (insertCast) {
	    w.write(")");
	}
    }

    /**
     * Translate a call to a method of an instantiated type.
     * This may require boxing/unboxing of primitives if the
     * type was instantiated on primitives.
     */
    protected void translateInstMethod(CodeWriter w, Translator tr,
            PolyJTypeSystem ts, Context c) {

        MethodInstance myMi = methodInstance();
        MethodInstance origMi = ts.originalMethod(myMi);
        if (origMi == null) {
            throw new InternalCompilerError("Found a method in an "
                + "instantiation of a polymorphic type that is not "
                + " in the original type!");
        }
        Type myReturn = myMi.returnType();
        Type origReturn = origMi.returnType();

        boolean origReturnIsParam = ts.isParameter(origReturn);
        boolean myReturnIsParam = ts.isParameter(myReturn);
        boolean isStmtExpr = ((PolyJExt) this.ext()).isValueIgnored();

        // Unbox the return value is it is a parameter to the
        // enclosing type, but we know it's a primitive here.
        boolean unboxReturn = origReturnIsParam && myReturn.isPrimitive();

        // Cast the return value when both the return type is a parameter
        // to the enclosing type, but we know the actual type here, and
        // the result of the method is used somewhere.
        boolean castReturn = origReturnIsParam && !(myReturnIsParam)
            && !isStmtExpr;

        writeInitialReturnCode(w, tr, ts, c, unboxReturn, castReturn, myReturn);
        PolyJTranslator.translateTargetNormal(w, tr, this, target);
        translateNameNormal(w, tr);
        PolyJTranslator.translateInstArgs(w, tr, ts, c, arguments(), 
            target.type().toClass(), myMi, origMi);
        writeFinalReturnCode(w, unboxReturn, castReturn);
    }

    /**
     * Translate the name of the method normally.
     */
    protected void translateNameNormal(CodeWriter w, Translator tr) {
        w.write(name);
    }

    /**
     * The default translation is to translate the target
     * and name normally, and to translate the args normally
     * except to add a where object if it's needed as the
     * first argument.
     */
    protected void translateDefault(CodeWriter w, Translator tr) {
        PolyJTranslator.translateTargetNormal(w, tr, this, target);
        translateNameNormal(w, tr);
        PolyJTranslator.translateArgs(w, tr, arguments(), methodInstance());
    }

    /**
     * Write the initial parentheses and cast of the return value,
     * if needed, for a call to a method in an instantiated class.
     */
    protected void writeInitialReturnCode(CodeWriter w, Translator tr,
            PolyJTypeSystem ts, Context c,
            boolean unboxReturn, boolean castReturn, Type myReturn) {
        if (unboxReturn) {
            PrimitiveType primitive = myReturn.toPrimitive();
            String primitiveWrapperName = ts.primitiveTypeWrapperName(primitive);
            w.write("((" + primitiveWrapperName + ")");
        } else if (castReturn) {
            w.write("((");
            if (myReturn instanceof InstType) {
                w.write(((InstType) myReturn).instType().translate(c));
            } else {
                w.write(myReturn.translate(c));
            }
            w.write(")");
        }
    }

    /**
     * Write the final unwrap and parentheses, if needed,
     * for a call to a method in an instantiated class.
     */
    protected void writeFinalReturnCode(CodeWriter w,
            boolean unboxReturn, boolean castReturn) {
        if (unboxReturn) {
            w.write(").Unwrap()");
        } else if (castReturn) {
            w.write(")");
        }
    }

}
