package polyglot.ext.polyj.ast;

import polyglot.ast.*;
import polyglot.types.*;
import polyglot.util.*;
import polyglot.visit.*;

import polyglot.ext.polyj.types.*;
import polyglot.ext.polyj.visit.*;

import polyglot.ext.jl.ast.*;

import java.util.*;

public class PolyJConstructorDecl_c extends ConstructorDecl_c 
    implements PolyJConstructorDecl 
{

    protected List constraints;
    
    /**
     * Constructors for PolyJMethodNodes with constraints
     */
    public PolyJConstructorDecl_c(Position pos, Flags flags, 
        String name, List formals, List throwTypes, List constraints, 
        Block body)
    {
        super(pos, flags, name, formals, throwTypes, body);
        this.constraints = TypedList.copyAndCheck(constraints, ClassConstraint.class, true);
    }

    protected PolyJConstructorDecl reconstruct(List formals, 
                                       List throwTypes, List constraints,
                                       Block body)
    {
        if (      !CollectionUtil.equals(this.formals, formals)
               || !CollectionUtil.equals(this.throwTypes, throwTypes)
               || !CollectionUtil.equals(this.constraints, constraints)
               || this.body != body) {
            PolyJConstructorDecl_c n = (PolyJConstructorDecl_c) super.reconstruct(formals, throwTypes, body);
            n.constraints = TypedList.copyAndCheck(constraints, ClassConstraint.class, true);
            return n;
        } else {
            return this;
        }
    }

    public List constraints() {
        return constraints;
    }

    public PolyJConstructorDecl constraints(List constraints) {
        PolyJConstructorDecl_c n = (PolyJConstructorDecl_c) copy();
        n.constraints = TypedList.copyAndCheck(constraints, ClassConstraint.class, true);
        return n;
    }

    public Node visitChildren(NodeVisitor v)
    {
        List newFormals = visitList(formals, v);
        List newExceptions = visitList(throwTypes, v);
        List newConstraints = visitList(constraints, v);
        Block newBody = (Block) visitChild(body, v);
        return reconstruct(newFormals, newExceptions, newConstraints, newBody);
    }

    public void dump(CodeWriter w)
    {
        w.write( "POLYJ CONSTRUCTOR");
        w.write( " < " + name + " >");
        w.write( " < " + flags.toString() + "> ");
    }

    public void prettyPrint(CodeWriter w, PrettyPrinter pp) {
        // FIXME: print constraints
        super.prettyPrint(w, pp);
    }

    /**
     * If the containing class of this constructor
     * carries a where-object, then we need to assign a value to
     * it from the argument we're passed.
     *
     * The exceptions are if the first statement in the body
     * of this constructor is a call to this(...), in
     * which case we don't do the assignment here.
     * Also if the first statement in the body is a call to
     * super(...), then we have to translate that call, and
     * then insert the assignment.
     */
    public void translate(CodeWriter w, Translator tr)
    {
        Context c = tr.context();
        PolyJTypeSystem ts = (PolyJTypeSystem) c.typeSystem();
        PolyJClassType curClass = (PolyJClassType) c.currentClassScope();
        boolean hasWhere = ts.needsWhereObject(curClass);

        translateHeader(w, tr, c, curClass, hasWhere);
        c = enterScope(c);
        tr = tr.context(c);

        boolean whereAssgn = false;
        boolean firstIsConstCall = false;
        boolean firstIsThisCall = false;
        
        List stmts = body().statements();
        if (!stmts.isEmpty()) {
            Stmt first = (Stmt) stmts.get(0);
            if (isSuperCall(first)) {
                firstIsConstCall = true;
                firstIsThisCall = false;
            } else if (isThisCall(first)) {
                firstIsConstCall = true;
                firstIsThisCall = true;
            }
        }
        
        if (hasWhere && !ts.inheritsWhereObject(curClass)) {
            if (stmts.isEmpty()) {
                whereAssgn = true;
            } else {
                Stmt first = (Stmt) stmts.get(0);
                if (isSuperCall(first)) {
                    whereAssgn = true;
                } else if (isThisCall(first)) {
                    whereAssgn = false;
                } else {
                    whereAssgn = true;
                }
            }
        }

        translateBody(w, tr, c, whereAssgn, firstIsConstCall, firstIsThisCall);
        w.newline(0);
    }

    /**
     * Returns whether s is a call to a superclass constructor.
     */
    protected boolean isSuperCall(Stmt s) {
        if (s instanceof ConstructorCall) {
            ConstructorCall c = (ConstructorCall) s;
            return (c.qualifier() == null)
                && (c.kind() == ConstructorCall.SUPER);
        } else {
            return false;
        }
    }

    /**
     * Returns whether s is a call to an alternate constructor.
     */
    protected boolean isThisCall(Stmt s) {
        if (s instanceof ConstructorCall) {
            ConstructorCall c = (ConstructorCall) s;
            return (c.qualifier() == null)
                && (c.kind() == ConstructorCall.THIS);
        } else {
            return false;
        }
    }

    /**
     * Translate the header of this constructor.  If we're
     * in a polymorphic class, that means we have to add
     * a where object as the first argument.
     */
    protected void translateHeader(CodeWriter w, Translator tr,
            Context c, PolyJClassType curClass, boolean inPolyClass) {
        w.begin(0);
        w.write(flags.translate());
        w.write(name);
        w.write("(");
        w.begin(0);

        // Add in where-object argument
        if (inPolyClass) {
            w.write(PolyJTranslator.adapterInterfaceShortName(curClass)
                + " $W");
            if (formals.iterator().hasNext()) {
                w.write(", ");
            }
        }

        for (Iterator iter = formals.iterator(); iter.hasNext(); )
        {
            Formal formal = (Formal) iter.next();
            formal.del().translate(w, tr);
            if (iter.hasNext()) {
                w.write( ",");
                w.allowBreak(0, " ");
            }
        }
        w.end();
        w.write( ")");

        if( !throwTypes.isEmpty()) {
            w.allowBreak(6);
            w.write( "throws ");
            for (Iterator i = throwTypes.iterator(); i.hasNext(); ) {
                TypeNode tn = (TypeNode) i.next();
                tn.del().translate(w, tr);
                if (i.hasNext()) {
                    w.write(","); w.allowBreak(4, " ");
                }
            }
        }
        w.end();
    }


    /**
     * Translate the body of this constructor.  If we're
     * in a polymorphic class, that means we may have have to add
     * a where object assignment.  We do so if addWhereAssgn is set.
     * We add the assignment after the first statement if afterFirst
     * is set, or as the first statement if it isn't.
     */
    protected void translateBody(CodeWriter w, Translator tr,
            Context c, boolean whereAssgn, 
            boolean firstIsConstCall, boolean firstIsThisCall)
    {
        Iterator it = body.statements().iterator();
        w.write(" ");
        c = body.enterScope(c);
        tr = tr.context(c);
        w.write("{");
        w.allowBreak(4," ");
        w.begin(0);

        // translate a superconstructor call
        if (firstIsConstCall) { //afterFirst) {
            Node n = (Node) it.next();
            printBlock(n, w, tr);
            if (it.hasNext()) {
                w.newline(0);
            }
        }

        // add the where object assignment
        if (whereAssgn) {
            w.begin(0);
            w.write("this.$W = $W;");
            w.end();
            if (it.hasNext()) {
                w.newline(0);
            }
        }

        if (!firstIsThisCall) {
            translateDefaultInit(w, (PolyJTranslator) tr, c);
            if (it.hasNext()) {
                w.newline(0);
            }
        }
        
        // finish the rest of the body
        for (;it.hasNext(); ) {
            Node n = (Node) it.next();
            printBlock(n, w, tr);
            if (it.hasNext()) {
                w.newline(0);
            }
        }
        w.end();
        w.allowBreak(0, " ");
        w.write("}");
    }
    
    protected void translateDefaultInit(CodeWriter w, PolyJTranslator tr, 
        Context c) 
    {
        PolyJTypeSystem ts = (PolyJTypeSystem) tr.typeSystem();
        ClassType scopedClass = c.currentClassScope();
        for (Iterator i = scopedClass.fields().iterator(); i.hasNext(); ) {
            FieldInstance fi = (FieldInstance) i.next();
            Type ft = fi.type();
            if (!fi.flags().isStatic() && ft instanceof ParamType) {
                ParamType container = (ParamType) ft;
                String name = fi.name();
                w.write("if (" + name + " == null) " + name + " = ");
                w.write("(("+ PolyJTranslator.adapterInterfaceFullName(scopedClass)
                    + ")" + tr.whereObjExpr() + ")." 
                    + ts.DefaultBinding((ParamType) container, scopedClass).name() 
                    + "()");
                w.write(";");
                w.newline(0);
            }
        }
    }
}
