package polyglot.ext.polyj.ast;

import polyglot.ext.polyj.types.*;
import polyglot.ext.polyj.util.*;
import polyglot.ext.polyj.visit.*;

import polyglot.ext.jl.ast.*;

import polyglot.ast.*;
import polyglot.types.*;
import polyglot.util.*;
import polyglot.visit.*;

import java.util.*;

/**
 * A PolyJNew is an immutable representation of the
 * use of the <code>new</code> operator to create a new instance of a class.
 * In addition to the type of the class being created, a
 * <code>PolyJNew</code> has a list of arguments to be passed to the
 * constructor of the object and an optional <code>ClassBody</code> used to
 * support anonymous classes. Such an expression may also be proceeded
 * by an qualifier expression which specifies the context in which the
 * object is being created.
 */
public class PolyJNew_c extends New_c implements PolyJNew {

    /**
     * Creates a new <code>PolyJNew</code>.
     *
     * @pre Requires that each element of <code>args</code> is an object of
     *      type <code>Expr</code>.
     */
    public PolyJNew_c( Position pos, Expr qualifier, TypeNode tn, List arguments, 
            ClassBody body) {
        super(pos, qualifier, tn, arguments, body);
    }

    protected TypeNode partialDisambTypeNode(TypeNode tn, TypeChecker tc, ClassType outer) throws SemanticException
    {
        if (tn instanceof InstTypeNode) {
            InstTypeNode itn = (InstTypeNode) tn;
            TypeNode base = super.partialDisambTypeNode(itn.base(), tc, outer);
            return itn.base(base);
        }
        else {
            return super.partialDisambTypeNode(tn, tc, outer);
        }
    }

    /**
     * Constructors of parametric classes require an additional operand
     * when translated, and where methods may have to be substituted
     * for type parameter constructors.
     */
    public void translate(CodeWriter w, Translator tr) {
        Context c = tr.context();
        PolyJTypeSystem ts = (PolyJTypeSystem) c.typeSystem();

        ConstructorInstance myCi = constructorInstance();
        ConstructorInstance origCi = ts.originalConstructor(myCi);

        if (tn.type() instanceof ParamType
                    && !ts.equals(myCi.container(), (ts.Object()))) {
            PolyJTranslator.translateWhereConstructor(w, tr, ts, c,
                (ParamType) tn.type(), myCi, arguments());
        } else if (ts.needsWhereObject(myCi.container())) {
            w.write( "new ");
            tn.del().translate(w, tr);
            PolyJTranslator.translateInstArgs(w, tr, ts, c, arguments(),
                tn.type().toClass(), myCi, origCi);
        } else {
            if (qualifier != null) {
                qualifier.del().translate(w, tr);
                w.write( ".");
            }

            w.write( "new ");
            tn.del().translate(w, tr);

            if (myCi.container() instanceof InstType) {
                PolyJTranslator.translateInstArgs(w, tr, ts, c, arguments(),
                    tn.type().toClass(), myCi, origCi);
            } else {
                PolyJTranslator.translateArgs(w, tr, arguments(), myCi);
            }

            if (body != null) {
                body.del().translate(w, tr);
            }
        }
    }
}
