package polyglot.ext.polyj.parse;

import polyglot.ext.polyj.ast.*;

import polyglot.ast.*;
import polyglot.visit.*;
import polyglot.types.*;
import polyglot.util.*;
import polyglot.parse.*;

import java.util.*;

/**
 * An <code>InstOrAccess</code> represents a <code>Parse</code> of the form "P[T]"
 * where T may be a type or an expression.  This could be either an instantiation
 * or an array access.  Both may be preceded by an ambiguous prefix.
 */
public class InstOrAccess extends Parse {
    Parse prefix;
    Parse param;

    public InstOrAccess(BaseParser parser, Position pos, Parse prefix, Parse param) throws Exception {
        super(parser, pos);
        this.prefix = prefix;
        this.param = param;

        if (prefix instanceof Array) parser.die("Illegal instantiation or array access.", pos);
    }

    public TypeNode toType() throws Exception {
        List l = new TypedList(new LinkedList(), ParamNode.class, false);
        l.add(param.toParamNode());
        return nf.InstTypeNode(pos, prefix.toType(), l);
    }

    public QualifierNode toQualifier() throws Exception {
        // The prefix must be a type.
        return toType();
    }
    
    public Prefix toPrefix() throws Exception { 
        return toReceiver(); 
    }

    public Receiver toReceiver() throws Exception {
        // The prefix must be either a type or an expression
        // (i.e., a receiver).
        return nf.AmbInstOrAccess(pos, prefix.toReceiver(),
                                         param.toReceiver());
    }

    public Expr toExpr() throws Exception {
        return nf.ArrayAccess(pos, prefix.toExpr(), param.toExpr());
    }

    public NewArray toNewArray(Position p, int addDims, ArrayInit init) throws Exception {
        Access a = new Access(parser, pos, prefix, param.toExpr());
        return a.toNewArray(p, addDims, init);
    }

    public Expr toNewArrayPrefix(Position p) throws Exception {
        // Only the first dimension of "new T[p][q][r]" is ambiguous;
        // the rest must be expressions.

        if (prefix instanceof Name) {
            // "new T.a[n]".  "name" may be either an expr or a param.
            List params = new LinkedList();
            return nf.AmbNewArray(p, prefix.toType(),
                                         param.toReceiver(),
                                         Collections.EMPTY_LIST);
        }
        else if (prefix instanceof Inst) {
            // "new T[L,M][n]".  "name" must be an expr.
            List dims = new LinkedList();
            dims.add(param.toExpr());
            return nf.NewArray(p, prefix.toType(), dims);
        }
        else if (prefix instanceof Access ||
                 prefix instanceof InstOrAccess) {
            // The prefix is "S[n]".  Recurse as if we were doing
            // "new S[n]".  "name" must be an expression.
            Expr e = prefix.toNewArrayPrefix(p);

            // Take the expression we built and add a new dimension.
            if (e instanceof NewArray) {
                NewArray a = (NewArray) e;
                List dims = new LinkedList(a.dims());
                dims.add(param.toExpr());
                return a.dims(dims);
            }
            else if (e instanceof AmbNewArray) {
                AmbNewArray a = (AmbNewArray) e;
                List dims = new LinkedList(a.dims());
                dims.add(param.toExpr());
                return a.dims(dims);
            }
        }

        parser.die(this.pos);
        return null;
    }
    
    public String toString() {
        return prefix + "[" + param + "]";    
    }

    public Parse prepend(Parse base) throws Exception {
        return new InstOrAccess(parser, pos, base, this);
    }
    
    
    public Receiver toParam() throws Exception {
        // An access can never be a parameter (since it cannot be final) 
        // thus this must be a type.
        return toType();
    }
    
}
